/*---------------------------------------------------------------------------*/
/* C++ encapsulation of AW V4 objects                                        */
/* Copyright (C) 2006-2007 Robert James Morales                              */
/* All rights reserved.                                                      */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions        */
/* are met:                                                                  */
/*                                                                           */
/*    - Redistributions of source code must retain the above copyright       */
/*      notice, this list of conditions and the following disclaimer.        */
/*                                                                           */
/*    - Redistributions in binary form must reproduce the above copyright    */
/*      notice, this list of conditions and the following disclaimer in      */
/*      the documentation and/or other materials provided with the           */
/*      distribution.                                                        */
/*                                                                           */
/*    - Neither the name of the author nor the names of its contributors     */
/*      may be used to endorse or promote products derived from this         */
/*      software without specific prior written permission.                  */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       */
/* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT         */
/* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS         */
/* FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE            */
/* COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,       */
/* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,      */
/* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS     */
/* OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED        */
/* AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,    */
/* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF     */
/* THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH      */
/* DAMAGE.                                                                   */
/*---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <memory>
#if defined(LINUX)
# include <netdb.h>
# include <sys/socket.h>
# include <fcntl.h>
# include <netinet/in.h>
# include <sys/time.h>
# include <arpa/inet.h>
# include <unistd.h>
#endif

#include "Aw.h"
#include "AwV4.h"

/*---------------------------------------------------------------------------*/

/** @brief Binary stream container.
 *
 *  An extremely simplistic binary stream container, used internally
 *  for string data contained in V4 objects
 */
class bstream
{
public:
	/** Add data to the stream
	 *  @param[in] buf Buffer pointing to the data to be added.
	 *  @param[in] bufsiz The size of the specified buffer.
	 */
	void Append(const void* buf, const size_t bufsiz)
	{
		if (buf != NULL && bufsiz != 0)
		{
			memcpy(m_stream + m_offset, buf, bufsiz);
			m_offset += bufsiz;
		}
	}

	/** Retrieve data from the stream
	 *  @param[in] bytes The number of bytes to retrieve.
	 *  @return An allocated buffer containing data. It is the caller's
	 *  responsibility to call delete[] on this buffer.
	 */
	template<typename T>
	T* Get(const size_t bytes)
	{
		if (bytes == 0)
			return NULL;

		T* buf_p = new T[bytes];
		memcpy(buf_p, m_stream + m_offset, bytes);

		m_offset += bytes;

		return buf_p;
	}

	/** Retrieve data from the stream as an AwV4::SimpleString object
	 *  @param[in] len The number of characters to retrieve
	 *  (not including the NUL terminator)
	 */
	AwV4::SimpleString GetString(const size_t len)
	{
		if (len == 0)
			return AwV4::SimpleString();

		std::auto_ptr<char> buf(new char[len+1]);

		char* buf_p = buf.get();

		memset(buf_p, 0, len+1);
		memcpy(buf_p, m_stream + m_offset, len);

		m_offset += len;

		return AwV4::SimpleString(buf_p);
	}

	/** Default constructor
	 *  @param[in,out] stream_p Pointer to an allocated buffer for the binary stream.
	 */
	bstream(char* stream_p)
		: m_offset(0), m_stream(stream_p)
	{
	}

protected:
	size_t m_offset;
	char* m_stream;

	/** Prevent copy-construction */
	bstream(const bstream&);
};

/*---------------------------------------------------------------------------*/

/** Initialize a 'aw_type_vector_range' structure */
static inline void InitVectorRange(aw_type_vector_range& vector)
{
	vector.min.x = 0;
	vector.min.y = 0;
	vector.min.z = 0;
	vector.max.x = 0;
	vector.max.y = 0;
	vector.max.z = 0;
}

/*---------------------------------------------------------------------------*/
/*                            A w V 4 : : Z o n e                            */
/*---------------------------------------------------------------------------*/

aw_object_data_zone* AwV4::Zone::GetObjectData()
{
	aw_object_data_zone* data = NULL;
	size_t strsiz = 0;

	/* Determine length of all strings for structure allocation */

	strsiz += m_footstep.Length();
	strsiz += m_ambient.Length();
	strsiz += m_camera.Length();
	strsiz += m_targetcur.Length();

	/* Allocate structure and initialize fields */

	if ((data = static_cast<aw_object_data_zone*>(malloc(sizeof(aw_object_data_zone) + strsiz))) == NULL)
		return NULL; // some kind of horrible memory failure

	memcpy(data, &m_data, sizeof(aw_object_data_zone));

	data->ambient_len = static_cast<unsigned short>(m_ambient.Length());
	data->camera_len = static_cast<unsigned char>(m_camera.Length());
	data->footstep_len = static_cast<unsigned short>(m_footstep.Length());
	data->target_cur_len = static_cast<unsigned char>(m_targetcur.Length());
	data->version = 0; // this is filled by the SDK

	memset(data->reserved, 0, sizeof(data->reserved));

	/* Build the string data */

	bstream bs(data->str_data);

	bs.Append(m_footstep.c_str(), data->footstep_len);
	bs.Append(m_ambient.c_str(), data->ambient_len);
	bs.Append(m_camera.c_str(), data->camera_len);
	bs.Append(m_targetcur.c_str(), data->target_cur_len);

	/* We're done! :D */

	return data;
}

/*---------------------------------------------------------------------------*/

AwV4::Zone::Zone()
{
	/* Default zone data */

	m_data.color = 0xFFFFFF;
	m_data.flags = 0;
	m_data.fog_max = 2500;
	m_data.fog_min = 1;
	m_data.friction = 1.0f;
	m_data.gravity = 1.0f;
	m_data.priority = 0;
	m_data.shape = AW_ZONE_TYPE_CUBE;
	m_data.size.x = 1.0f;
	m_data.size.y = 1.0f;
	m_data.size.z = 1.0f;
}

/*---------------------------------------------------------------------------*/

void AwV4::Zone::Assign(const aw_object_data_zone* data)
{
	m_ambient.Free();
	m_camera.Free();
	m_footstep.Free();
	m_targetcur.Free();

	memcpy(&m_data, data, sizeof(aw_type_vector));

	bstream bs(const_cast<char*>(data->str_data));

	if (data->footstep_len != 0)
		m_footstep = bs.GetString(data->footstep_len);
	if (data->ambient_len != 0)
		m_ambient = bs.GetString(data->ambient_len);
	if (data->camera_len != 0)
		m_camera = bs.GetString(data->camera_len);
	if (data->target_cur_len != 0)
		m_targetcur = bs.GetString(data->target_cur_len);
}

/*---------------------------------------------------------------------------*/
/*                 A w V 4 : : P a r t i c l e E m i t t e r                 */
/*---------------------------------------------------------------------------*/

aw_object_data_particles* AwV4::ParticleEmitter::GetObjectData()
{
	aw_object_data_particles* data = NULL;
	size_t strsiz = 0;

	/* Determine length of all strings for structure allocation */

	strsiz += m_assets.Length();
	strsiz += m_tagname.Length();

	/* Allocate structure and initialize fields */

	if ((data = static_cast<aw_object_data_particles*>(malloc(sizeof(aw_object_data_particles) + strsiz))) == NULL)
		return NULL; // some kind of horrible memory failure

	memcpy(data, &m_data, sizeof(aw_object_data_particles));

	data->asset_list_len = static_cast<unsigned short>(m_assets.Length());
	data->name_len = static_cast<unsigned char>(m_tagname.Length());
	data->version = 0; // this is filled by the SDK

	memset(data->reserved, 0, sizeof(data->reserved));

	/* Build the string data */

	bstream bs(data->str_data);

	bs.Append(m_assets.c_str(), data->asset_list_len);
	bs.Append(m_tagname.c_str(), data->name_len);

	/* We're done! :D */

	return data;
}

/*---------------------------------------------------------------------------*/

AwV4::ParticleEmitter::ParticleEmitter()
{
	m_data.release_min = 0;
	m_data.release_max = 0;
	m_data.release_size = 0;
	m_data.lifespan = 0;
	m_data.emitter_lifespan = 0;
	m_data.fade_in = 0;
	m_data.fade_out = 0;
	m_data.color_start = 0;
	m_data.color_end = 0;
	m_data.opacity = 1.0f;
	m_data.render_style = AW_PARTICLE_DRAW_NORMAL;
	m_data.flags = 0;
	m_data.style = AW_PARTICLE_TYPE_SPRITE;
	InitVectorRange(m_data.volume);
	InitVectorRange(m_data.speed);
	InitVectorRange(m_data.accel);
	InitVectorRange(m_data.angle);
	InitVectorRange(m_data.spin);
	InitVectorRange(m_data.size);
}

/*---------------------------------------------------------------------------*/

void AwV4::ParticleEmitter::Assign(const aw_object_data_particles* data)
{
	m_assets.Free();
	m_tagname.Free();

	memcpy(&m_data, data, sizeof(aw_object_data_particles));

	bstream bs(const_cast<char*>(data->str_data));

	if (data->asset_list_len != 0)
		m_assets = bs.GetString(data->asset_list_len);
	if (data->name_len != 0)
		m_tagname = bs.GetString(data->name_len);
}

/*---------------------------------------------------------------------------*/
/*                          A w V 4 : : C a m e r a                          */
/*---------------------------------------------------------------------------*/

aw_object_data_camera* AwV4::Camera::GetObjectData()
{
	aw_object_data_camera* data = NULL;
	size_t strsiz = 0;

	/* Determine length of all strings for structure allocation */

	strsiz += m_tag.Length();

	/* Allocate structure and initialize fields */

	if ((data = static_cast<aw_object_data_camera*>(malloc(sizeof(aw_object_data_camera) + strsiz))) == NULL)
		return NULL; // some kind of horrible memory failure

	memcpy(data, &m_data, sizeof(aw_object_data_camera));

	data->name_len = static_cast<unsigned char>(m_tag.Length());
	data->version = 0; // this is filled by the SDK

	/* Build the string data */

	bstream bs(data->str_data);
	bs.Append(m_tag.c_str(), data->name_len);

	/* We're done! :D */

	return data;
}

/*---------------------------------------------------------------------------*/

AwV4::Camera::Camera()
{
	m_data.flags = 0;
	m_data.zoom = 0.0f;
}

/*---------------------------------------------------------------------------*/

void AwV4::Camera::Assign(const aw_object_data_camera* data)
{
	m_tag.Free();

	memcpy(&m_data, data, sizeof(aw_object_data_camera));

	bstream bs(const_cast<char*>(data->str_data));

	if (data->name_len != 0)
		m_tag = bs.GetString(data->name_len);
}

/*---------------------------------------------------------------------------*/
/*                           A w V 4 : : M o v e r                           */
/*---------------------------------------------------------------------------*/

void AwV4::Mover::AddWaypoint(Waypoint* wp)
{
	if (m_waypoint != NULL)
	{
		Waypoint* last_p = NULL;
		for (last_p = m_waypoint; last_p->m_next != NULL; last_p = last_p->m_next)
			;

		last_p->m_next = wp;
	}
	else
		m_waypoint = wp;

	m_wpcount++;
}

/*---------------------------------------------------------------------------*/

void AwV4::Mover::RemoveWaypoint(Waypoint*& wp)
{
	if (wp == NULL)
		return;

	Waypoint* sav_p = NULL;

	if (wp == m_waypoint)
	{
		sav_p = wp->m_next;

		delete wp;
		m_wpcount--;

		m_waypoint = sav_p;
	}
	else
	{
		for (Waypoint* ptr = m_waypoint; ptr->m_next != NULL; ptr = ptr->m_next)
		{
			if (ptr->m_next == wp)
			{
				sav_p = ptr->m_next->m_next;

				delete ptr->m_next;
				m_wpcount--;

				ptr->m_next = sav_p;
			}
		}
	}

	wp = NULL;
}

/*---------------------------------------------------------------------------*/

void AwV4::Mover::ClearWaypoints()
{
	Waypoint* ptr = m_waypoint;
	Waypoint* sav_p = NULL;
	
	while (ptr != NULL)
	{
		sav_p = ptr->m_next;
		delete ptr;

		ptr = sav_p;
	}

	m_waypoint = 0;
	m_wpcount = 0;
}

/*---------------------------------------------------------------------------*/

aw_object_data_mover* AwV4::Mover::GetObjectData()
{
	aw_object_data_mover* data = NULL;
	size_t strsiz = 0;
	aw_waypoint* wpbuf = new aw_waypoint[m_wpcount];

	/* Determine length of all strings for structure allocation */

	strsiz += m_name.Length();
	strsiz += m_seq.Length();
	strsiz += m_script.Length();
	strsiz += m_sound.Length();
	strsiz += sizeof(aw_waypoint) * m_wpcount;
	strsiz += m_bumpname.Length();

	size_t wx = 0;
	for (Waypoint* wp = m_waypoint; wp != NULL; wp = wp->GetNext())
	{
		wp->Assign(&wpbuf[wx++]);
	}

	/* Allocate structure and initialize fields */

	if ((data = static_cast<aw_object_data_mover*>(malloc(sizeof(aw_object_data_mover) + strsiz))) == NULL)
		return NULL; // some kind of horrible memory failure

	memcpy(data, &m_data, sizeof(aw_object_data_mover));

	data->name_len = static_cast<unsigned char>(m_name.Length());
	data->seq_len = static_cast<unsigned char>(m_seq.Length());
	data->script_len = static_cast<unsigned char>(m_script.Length());
	data->sound_len = static_cast<unsigned char>(m_sound.Length());
	data->waypoints_len = static_cast<unsigned short>(sizeof(aw_waypoint) * m_wpcount);
	data->bump_name_len = static_cast<unsigned char>(m_bumpname.Length());
	data->version = 0; // this is filled by the SDK

	/* Build the string data */

	bstream bs(data->str_data);

	bs.Append(m_name.c_str(), data->name_len);
	bs.Append(m_seq.c_str(), data->seq_len);
	bs.Append(m_script.c_str(), data->script_len);
	bs.Append(m_sound.c_str(), data->sound_len);

	for (wx = 0; wx < m_wpcount; wx++)
		bs.Append(&wpbuf[wx], sizeof(aw_waypoint));

	bs.Append(m_bumpname.c_str(), data->bump_name_len);

	/* We're done! :D */

	return data;
}

/*---------------------------------------------------------------------------*/

AwV4::Mover::Mover()
	: m_waypoint(NULL), m_wpcount(0)
{
	memset(&m_data, 0, sizeof(aw_object_data_mover));

	/* Mover defaults */
	m_data.type = static_cast<unsigned char>(AW_MOVER_TYPE_ACTIVE);
	m_data.flags = AW_MOVER_FLAG_SINGLE_RIDER|AW_MOVER_FLAG_CLICK_START|AW_MOVER_FLAG_CLICK_PAUSE|AW_MOVER_FLAG_EMPTY_RESET|AW_MOVER_FLAG_LOCK_PITCH;
	m_data.speed_factor = static_cast<unsigned char>(1.0f * 10);
	m_data.turn_factor = static_cast<unsigned char>(1.0f * 10);
	m_data.friction_factor = static_cast<unsigned char>(1.0f * 10);
}

/*---------------------------------------------------------------------------*/

void AwV4::Mover::Assign(const aw_object_data_mover* recv)
{
	/* Clear all string fields */

	m_name.Free();
	m_seq.Free();
	m_script.Free();
	m_sound.Free();
	m_bumpname.Free();

	ClearWaypoints();

	/* Copy basic data */

	memcpy(&m_data, recv, sizeof(aw_object_data_mover));

	/* Copy all string fields and waypoints */

	bstream bs(const_cast<char*>(recv->str_data));

	if (recv->name_len != 0)
		m_name = bs.GetString(recv->name_len);
	if (recv->seq_len != 0)
		m_seq = bs.GetString(recv->seq_len);
	if (recv->script_len != 0)
		m_script = bs.GetString(recv->script_len);
	if (recv->sound_len != 0)
		m_sound = bs.GetString(recv->sound_len);

	if (recv->waypoints_len != 0)
	{
		aw_waypoint* wpbuf = bs.Get<aw_waypoint>(recv->waypoints_len);
		size_t count = static_cast<size_t>(recv->waypoints_len) / sizeof(aw_waypoint);

		for (size_t wx = 0; wx < count; wx++)
			AddWaypoint(new AwV4::Mover::Waypoint(wpbuf[wx]));

		delete[] wpbuf;
	}

	if (recv->bump_name_len != 0)
		m_bumpname = bs.GetString(recv->bump_name_len);
}

/*---------------------------------------------------------------------------*/
/*                       G e n e r a l   M e t h o d s                       */
/*---------------------------------------------------------------------------*/

void AwV4::PrepareToBuild(aw_object_data_zone* zone_p)
{
	unsigned int bsiz = sizeof(aw_object_data_zone);

	bsiz += zone_p->footstep_len;
	bsiz += zone_p->ambient_len;
	bsiz += zone_p->camera_len;
	bsiz += zone_p->target_cur_len;

	aw_string_set(AW_OBJECT_DESCRIPTION, "");
	aw_string_set(AW_OBJECT_ACTION, "");
	aw_string_set(AW_OBJECT_MODEL, "");
	aw_int_set(AW_OBJECT_TYPE, AW_OBJECT_TYPE_ZONE);
	aw_data_set(AW_OBJECT_DATA, reinterpret_cast<char*>(zone_p), bsiz);
}

/*---------------------------------------------------------------------------*/

void AwV4::PrepareToBuild(aw_object_data_particles* particle_p)
{
	unsigned int bsiz = sizeof(aw_object_data_particles);

	bsiz += particle_p->asset_list_len;
	bsiz += particle_p->name_len;

	aw_string_set(AW_OBJECT_DESCRIPTION, "");
	aw_string_set(AW_OBJECT_ACTION, "");
	aw_string_set(AW_OBJECT_MODEL, "");
	aw_int_set(AW_OBJECT_TYPE, AW_OBJECT_TYPE_PARTICLES);
	aw_data_set(AW_OBJECT_DATA, reinterpret_cast<char*>(particle_p), bsiz);
}

/*---------------------------------------------------------------------------*/

void AwV4::PrepareToBuild(aw_object_data_camera* camera_p)
{
	unsigned int bsiz = sizeof(aw_object_data_camera);

	bsiz += camera_p->name_len;

	aw_string_set(AW_OBJECT_DESCRIPTION, "");
	aw_string_set(AW_OBJECT_ACTION, "");
	aw_string_set(AW_OBJECT_MODEL, "");
	aw_int_set(AW_OBJECT_TYPE, AW_OBJECT_TYPE_PARTICLES);
	aw_data_set(AW_OBJECT_DATA, reinterpret_cast<char*>(camera_p), bsiz);
}

/*---------------------------------------------------------------------------*/

void AwV4::PrepareToBuild(aw_object_data_mover* mover_p)
{
	unsigned int bsiz = sizeof(aw_object_data_mover);

	bsiz += mover_p->bump_name_len;
	bsiz += mover_p->name_len;
	bsiz += mover_p->script_len;
	bsiz += mover_p->seq_len;
	bsiz += mover_p->sound_len;
	bsiz += mover_p->waypoints_len;

	aw_string_set(AW_OBJECT_DESCRIPTION, "");
	aw_string_set(AW_OBJECT_ACTION, "");
	aw_int_set(AW_OBJECT_TYPE, AW_OBJECT_TYPE_MOVER);
	aw_data_set(AW_OBJECT_DATA, reinterpret_cast<char*>(mover_p), bsiz);
}

/*---------------------------------------------------------------------------*/
