/*---------------------------------------------------------------------------*/
/* C++ encapsulation of AW V4 objects                                        */
/* Copyright (C) 2007 Robert James Morales                                   */
/* All rights reserved.                                                      */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions        */
/* are met:                                                                  */
/*                                                                           */
/*    - Redistributions of source code must retain the above copyright       */
/*      notice, this list of conditions and the following disclaimer.        */
/*                                                                           */
/*    - Redistributions in binary form must reproduce the above copyright    */
/*      notice, this list of conditions and the following disclaimer in      */
/*      the documentation and/or other materials provided with the           */
/*      distribution.                                                        */
/*                                                                           */
/*    - Neither the name of the author nor the names of its contributors     */
/*      may be used to endorse or promote products derived from this         */
/*      software without specific prior written permission.                  */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       */
/* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT         */
/* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS         */
/* FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE            */
/* COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,       */
/* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,      */
/* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS     */
/* OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED        */
/* AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,    */
/* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF     */
/* THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH      */
/* DAMAGE.                                                                   */
/*---------------------------------------------------------------------------*/
/* Updated: 04 MAR 2007                                                      */
/*---------------------------------------------------------------------------*/

#ifndef _AWV4_H__
#define _AWV4_H__
/*---------------------------------------------------------------------------*/

/** Designed for AW SDK build 65 */
#define AWV4_SDK_BUILD (65)

#if AWV4_SDK_BUILD < AW_BUILD
# error Please obtain an updated version of this wrapper library.
#elif AWV4_SDK_BUILD > AW_BUILD
# error Please obtain an updated version of the SDK.
#endif

/*---------------------------------------------------------------------------*/

/** @brief Methods and classes to handle AW 4.1 objects
 *
 *  This namespace contains all the encapsulation classes, utility classes,
 *  and methods necessary for manipulating 4.1 object structures.
 */
namespace AwV4
{
	/** @brief An extremely primitive string class.
	 *
	 *  This class implements only the bare minimum required for the
	 *  V4 class wrappers.
	 *
	 *  Reference counting is implemented to reduce memory usage; it is
	 *  therefore not recommended you do any funny business with the
	 *  pointer returned by c_str()
	 */
	class SimpleString
	{
	public:
		/** Returns a C-style compatible string; funny business not advised!
		 *
		 *  If you need to do anything complicated, just copy the data and
		 *  do it yourself (perhaps with another more sophisticated string
		 *  library)
		 */
		const char* c_str() { return (m_data != NULL && m_data->refcnt != 0)? m_data->str : NULL; }

		/** Determine the length of this string, not including the NUL
		 *  terminator.
		 */
		size_t Length() const { return (m_data != NULL && m_data->refcnt != 0)? (m_data->size - 1) : 0; }

		/** Assign data to this string. */
		void Assign(const char* str_p)
		{
			if (str_p != NULL)
			{
				size_t len = strlen(str_p) + 1;
				init(len);
				memmove(m_data->str, str_p, len);
			}
			else
				Free();
		}

		/** Add data to the end of this string */
		void Append(const char* str_p)
		{
			if (str_p == NULL)
				return;

			if (m_data != NULL)
			{
				size_t oldlen = m_data->size - 1;
				size_t newlen = strlen(str_p);
				init(m_data->size + newlen);
				memmove(m_data->str + oldlen, str_p, newlen);
			}
			else
				Assign(str_p);
		}

		/** Default constructor */
		SimpleString(const char* str_p = NULL)
			: m_data(NULL)
		{
			Assign(str_p);
		}

		/** Copy constructor */
		SimpleString(const SimpleString& rhs) : m_data(NULL) { operator=(rhs); }

		/** Assignment operator */
		const SimpleString& operator=(const SimpleString& rhs)
		{
			// just copy the pointer; we'll do copy-on-write if we need
			// to change data

			if (rhs.m_data != NULL && m_data != rhs.m_data)
			{
				quit();

				m_data = const_cast<StringData*>(rhs.m_data);
				m_data->refcnt++;
			}
			else if (rhs.m_data == NULL)
				quit();

			return *this;
		}

		/** Alternate assignment operator, accepting a C-string */
		SimpleString& operator=(const char* rhs)
		{
			Assign(rhs);
			return *this;
		}

		/** Default destructor */
		~SimpleString() { quit(); }

		/** Free the contents of the string */
		void Free() { quit(); }

	protected:

#pragma pack (push, AWV4_SIMPLESTRING, 1)        // Store current structure packing alignment
#pragma pack (1)
		/** Internal string representation */
		struct StringData
		{
			size_t size;
			size_t refcnt;
			char str[1];
		};
#pragma pack (pop, AWV4_SIMPLESTRING)            // Restore structure packing alignment

		StringData* m_data;

		/** Allocate space for the string */
		void init(size_t siz)
		{
			if (siz == 0)
				return;

			// Allocate a new structure if we haven't already done so, or if
			// this is not our pointer (refcount is higher than 1; copy-on-write)
			if (m_data == NULL || m_data->refcnt > 1)
			{
				StringData* data_p = NULL;

				data_p = static_cast<StringData*>(malloc(sizeof(StringData) + siz));
				data_p->refcnt = 1;

				if (m_data != NULL)
				{
					m_data->refcnt--;
					memmove(data_p->str, m_data->str, m_data->size);
				}

				m_data = data_p;
			}
			else
				m_data = static_cast<StringData*>(realloc(m_data, sizeof(StringData) + siz));

			m_data->str[ m_data->size = siz ] = '\0';
		}

		/** Deallocate string memory */
		void quit()
		{
			if (m_data == NULL)
				return;

			m_data->refcnt--;

			if (m_data->refcnt == 0)
			{
				free(m_data);
			}

			m_data = NULL;
		}
	};

	/** @brief A wrapper class defining a 'zone' object type.
	 *
	 * This class encapsulates a zone object. The methods return
	 * references to various fields, which may be used to update them.
	 *
	 * For example:
	 *
	 *     float& zgrav = myzone.Gravity();
	 *
	 * Will allow you to update the zone's gravity by assigning a value
	 * to the 'zgrav' reference.
	 */
	class Zone
	{
	public:
		/** @brief Determine this zone's size.
		 *
		 *  This defines the size of the zone.
		 *  See "Volume Shape" above for details on how these values
		 *  are used.
		 */
		aw_type_vector& Size() { return m_data.size; }

		/** @brief Determine the shape of this zone.
		 *
		 *  This controls the shape of the area. There are 3 types:
		 *
		 *    1) AW_ZONE_TYPE_CUBE - The zone is a rectangular area in
		 *       size as defined by the Size values.
		 *
		 *    2) AW_ZONE_TYPE_CYLINDER - The zone is a cylindrical area.
		 *       The X Size defines the radius of the cylinder, the
		 *       Y Size defines the height, and the Z Size is unused.
		 *
		 *    3) AW_ZONE_TYPE_SPHERE - The zone is a sphere whose
		 *       radius is defined by the X Size value. The Y and
		 *       Z values are not used.
		 */
		unsigned char& Shape() { return m_data.shape; }

		/** @brief Determine this zone's priority value
		 *
		 *  When two zones intersect, there needs to be some way to
		 *  define which zone supersedes the other. When you stand in a
		 *  spot where two zones intersect, which zone are you really
		 *  in? The priority value defines how important this zone is
		 *  in relation to others. The lower the number, the more
		 *  important the zone is.
		 */
		char& Priority() { return m_data.priority; }

		/** @brief Determine the gravity of this zone.
		 *
		 * This controls how fast users will fall when in this zone.
		 * A value of 1.00 is "normal" gravity.
		 */
		float& Gravity() { return m_data.gravity; }

		/** @brief Determine the friction value in this zone.
		 *
		 *  This controls how fast users will stop when walking in this
		 *  zone. A value of 1.00 is considered "normal" friction.
		 *  Lower values will make walking "slippery", while higher
		 *  values will slow walking speed and cause the user to stop
		 *  more quickly.
		 */
		float& Friction() { return m_data.friction; }

		/** @brief Return the bitfield containing zone options.
		 *
		 *  This value is a bitfield which contains all of the active
		 *  zone options.
		 *
		 *  Zone flags are defined in AW_ZONE_FLAGS. Examples follow.
		 *
		 *  To determine if the 'AW_ZONE_FLAG_VISIBLE' flag is set:
		 *
		 *    bool bIsZoneVisible = (myzone.Flags() & AW_ZONE_FLAG_VISIBLE)? true : false;
		 *
		 *  To remove a flag:
		 *
		 *    unsigned short& opts = myzone.Flags();
		 *    opts &= ~AW_ZONE_FLAG_VISIBLE;
		 *
		 *  To add a flag:
		 *
		 *    unsigned short& opts = myzone.Flags();
		 *    opts |= AW_ZONE_FLAG_VISIBLE;
		 */
		unsigned short& Flags() { return m_data.flags; }

		/** @brief Determine the fog and water color for this zone.
		 *
		 *  This controls the color of the fog in this zone, and the
		 *  color of the water (if applicable).
		 */
		unsigned long& Color() { return m_data.color; }

		/** @brief Determine the zone's 'near' fog setting. */
		unsigned short& FogMin() { return m_data.fog_max; }

		/** @brief Determine the zone's 'far' fog setting. */
		unsigned short& FogMax() { return m_data.fog_min; }

		/** @brief Determine this zone's footstep sound, if any.
		 *
		 *  This sound is used if the avatar's feet are in this zone
		 *  while walking or running.
		 *
		 *  If no sound is specified, NULL is returned.
		 */
		SimpleString& FootstepSound() { return m_footstep; }

		/** @brief Determine this zone's ambient sound, if any.
		 *
		 *  This sound (if specified) will repeat in the background
		 *  when the camera (usually the avatar's head) is in this
		 *  zone.
		 *
		 *  If no sound is specified, NULL is returned.
		 */
		SimpleString& AmbientSound() { return m_ambient; }

		/** @brief Determine this zone's camera tag, if any.
		 *
		 *  This specifies that name of a particular camera object
		 *  within the scene. When the user enters this zone, their
		 *  camera will switch to the camera matching the given name,
		 *  and will return to the previous setting when they leave.
		 *
		 *  If no tag is specified, NULL is returned.
		 */
		SimpleString& CameraTag() { return m_camera; }

		/** @brief Determine this zone's target cursor, if any.
		 *
		 * This specifies the texture that will be used as the center
		 * cursor in Move Mode (mouselook).
		 *
		 *  If no cursor is specified, NULL is returned.
		 */
		SimpleString& TargetCursor() { return m_targetcur; }

		/** @brief Create an aw_object_data_zone structure
		 *
		 *  Creates and returns an aw_object_data_zone structure
		 *  using the data from this wrapper class.
		 *
		 *  Note that it is up to the caller to appropriately use
		 *  free() to delete the memory associated with this
		 *  structure.
		 */
		aw_object_data_zone* GetObjectData();

		/** @brief Default constructor.
		 *
		 *  Constructor which initializes the zone wrapper class.
		 */
		Zone();

		/** @brief Assign zone data from an SDK structure.
		 *
		 *  Constructor which initializes the zone wrapper class from
		 *  an existing aw_object_data_zone structure.
		 */
		void Assign(const aw_object_data_zone* data);

	private:
		aw_object_data_zone m_data;
		SimpleString m_footstep;
		SimpleString m_ambient;
		SimpleString m_camera;
		SimpleString m_targetcur;

		/** Prevention of copy-construction */
		Zone(const Zone&);
	};

	/** @brief A wrapper class defining a 'particle emitter' object.
	 *
	 * This class encapsulates a particle emitter. The methods return
	 * references to various fields, which may be used to update them.
	 *
	 * For example:
	 *
	 *     float& opac = emitter.Opacity();
	 *
	 * Will allow you to update the emitter's opacity by assigning a
	 * value to the 'opac' reference.
	 */
	class ParticleEmitter
	{
	public:
		/** @brief Determine the emitter's effective volume.
		 *
		 *  This defines the size (in meters) of the cubic area in
		 *  which new particles will appear.
		 */
		aw_type_vector_range& Volume() { return m_data.volume; }

		/** @brief Determine the emitter's speed.
		 *
		 *  This defines how fast (in meters per second) particles will
		 *  be moving at the time of their creation.
		 */
		aw_type_vector_range& Speed() { return m_data.speed; }

		/** @brief Determine the emitter's acceleration.
		 *
		 *  This defines how fast (in meters per second) particles will
		 *  accelerate.
		 */
		aw_type_vector_range& Accel() { return m_data.accel; }

		/** @brief Determine the emitter's starting angle.
		 *
		 *  This defines the position (in degrees of rotation) of each
		 *  particle.
		 */
		aw_type_vector_range& Angle() { return m_data.angle; }

		/** @brief Determine the emitter's rotation.
		 *
		 *  This defines the rotation (in degrees of rotation per
		 *  second) of each particle.
		 */
		aw_type_vector_range& Spin() { return m_data.spin; }

		/** @brief Determine the emitter's size.
		 *
		 *  This defines how big (in meters) each particle is.
		 *  These numbers have different effects based on what type
		 *  of particle this is.
		 */
		aw_type_vector_range& Size() { return m_data.size; }

		/** @brief Determine the emitter's release time.
		 *
		 *  This defines how often (in milliseconds) new particles will
		 *  appear. If this range is set to 500 to 3000, then particles
		 *  will be released at least once every three seconds, but not
		 *  more often than once every half second. The browser will
		 *  emit particles at random intervals between these two
		 *  values. 
		 */
		unsigned long& ReleaseMin() { return m_data.release_min; }

		/** @see ReleaseMin */
		unsigned long& ReleaseMax() { return m_data.release_max; }

		/** @brief Determine the particle release count.
		 *
		 *  This defines how many particles will be released at one time.
		 */
		unsigned short& ReleaseCount() { return m_data.release_size; }

		/** @brief Determine the particle lifespan.
		 *
		 *  This defines how long (in milliseconds) each particle will
		 *  last once it is released.
		 */
		unsigned long& Lifespan() { return m_data.lifespan; }

		/** @brief Determine the emitter's lifespan.
		 *
		 *  This defines how long (in milliseconds) this emitter will
		 *  release particles.
		 */
		unsigned long& EmitterLifespan() { return m_data.emitter_lifespan; }

		/** @brief Determine the particle fade-in time.
		 *
		 *  This defines how long (in milliseconds) it will take for
		 *  each particle to "fade in" to view.
		 */
		unsigned long& FadeIn() { return m_data.fade_in; }

		/** @brief Determine the particle fade-out time.
		 *
		 *  This defines how long (in milliseconds) it will take for
		 *  each particle to "fade out" of view.
		 */
		unsigned long& FadeOut() { return m_data.fade_out; }

		/** @brief Determine the particle start color.
		 *
		 *  If "Interpolate" is enabled, then particles will begin at
		 *  this color and fade to "Color End". If it is not enabled,
		 *  then particles will randomly choose a color between
		 *  "Color Start" and "Color End".
		 */
		unsigned long& ColorStart() { return m_data.color_start; }

		/** @brief Determine the particle end color.
		 *
		 *  If "Interpolate" is enabled, then particles will begin at
		 *  "Color Start" and fade to this color. If it is not enabled,
		 *  then particles will randomly choose a color between
		 *  "Color Start" and "Color End".
		 */
		unsigned long& ColorEnd() { return m_data.color_end; }

		/** @brief Determine the particle opacity.
		 *
		 *  This is a value between 0.00 and 1.00 that defines how
		 *  opaque the particles will be.
		 */
		float& Opacity() { return m_data.opacity; }

		/** @brief Determine the particle render style.
		 *
		 *  This defines how the particle will be blended with the scene.
		 *
		 *    1) AW_PARTICLE_DRAW_NORMAL: The texture is simply drawn
		 *       to the scene with no special effects.
		 *
		 *    2) AW_PARTICLE_DRAW_BRIGHT: The texture is drawn to the
		 *       scene, blending with whatever is already visible so as
		 *       to brighten the scene.
		 *
		 *    3) AW_PARTICLE_DRAW_GLOW: The texture is added to the
		 *       scene, making it significantly brighter.
		 */
		unsigned char& RenderStyle() { return m_data.render_style; }

		/** @brief Return the bitfield containing emitter options.
		 *
		 *  This value is a bitfield which contains all of the active
		 *  emitter options.
		 *
		 *  Emitter flags are defined in AW_PARTICLE_FLAGS. Examples follow.
		 *
		 *  To determine if the 'AW_PARTICLE_FLAG_INTERPOLATE' flag is set:
		 *
		 *    bool bInterpolate = (emitter.Flags() & AW_PARTICLE_FLAG_INTERPOLATE)? true : false;
		 *
		 *  To remove a flag:
		 *
		 *    unsigned short& opts = emitter.Flags();
		 *    opts &= ~AW_PARTICLE_FLAG_INTERPOLATE;
		 *
		 *  To add a flag:
		 *
		 *    unsigned short& opts = emitter.Flags();
		 *    opts |= AW_PARTICLE_FLAG_INTERPOLATE;
		 */
		unsigned short& Flags() { return m_data.flags; }

		/** @brief Determine the type of particles to release.
		 *
		 *  This controls what kind of particle will be released.
		 *  There are four types:
		 *
		 *    1) AW_PARTICLE_TYPE_SPRITE
		 *    2) AW_PARTICLE_TYPE_FACER
		 *    3) AW_PARTICLE_TYPE_FLAT
		 *    4) AW_PARTICLE_TYPE_OBJECT
		 */
		unsigned char& Type() { return m_data.style; }

		/** @brief Determine the emitter's asset list, if any.
		 *
		 *  This defines what textures or models will be used to make
		 *  the particles. This may be a list of textures
		 *  (seperated by commas) for sprites, facers, and flat panels,
		 *  or a list of model names (seperated by commas) for model
		 *  type particles.
		 */
		SimpleString& AssetList() { return m_assets; }

		/** @brief Determine the emitter's tag name, if any.
		 *
		 *  This assigns a name to the emitter. If this field is NOT
		 *  empty, the emitter will NOT start when it comes into view,
		 *  but will need to be triggered by another object.
		 */
		SimpleString& TagName() { return m_tagname; }

		/** @brief Create an aw_object_data_particles structure
		 *
		 *  Creates and returns an aw_object_data_particles structure
		 *  using the data from this wrapper class.
		 *
		 *  Note that it is up to the caller to appropriately use
		 *  free() to delete the memory associated with this
		 *  structure.
		 */
		aw_object_data_particles* GetObjectData();

		/** @brief Default constructor.
		 *
		 *  Constructor which initializes the emitter wrapper class.
		 */
		ParticleEmitter();

		/** @brief Assign emitter data from an SDK structure.
		 *
		 *  Constructor which initializes the zone wrapper class from
		 *  an existing aw_object_data_particles structure.
		 */
		void Assign(const aw_object_data_particles* data);

		/** @brief Determine the efficiency of an emitter.
		 *
		 *  Calculates an efficiency rating for this emitter.
		 *  For more info, see:
		 *  http://www.swcity.net/academy/index.php?n=Main.ParticleEmitterOptimization
		 */
		unsigned int Efficiency() const
		{
			int rel_time = (m_data.release_min != m_data.release_max)? (m_data.release_min + m_data.release_max) / 2 : m_data.release_min;
			return (m_data.lifespan / rel_time) * m_data.release_size;
		}

	protected:
		aw_object_data_particles m_data;
		SimpleString m_tagname;
		SimpleString m_assets;

		ParticleEmitter(const ParticleEmitter&);
	};

	/** @brief A wrapper class defining a 'camera' object.
	 *
	 * This class encapsulates a camera object. The methods return
	 * references to various fields, which may be used to update them.
	 *
	 * For example:
	 *
	 *     float& zoom = cam.Zoom();
	 *
	 * Will allow you to update the camera's zoom by assigning a
	 * value to the 'zoom' reference.
	 */
	class Camera
	{
	public:
		/** @brief Return the bitfield containing camera options.
		 *
		 *  This value is a bitfield which contains all of the active
		 *  camera options.
		 *
		 *  Camera flags are defined in AW_CAMERA_FLAGS. Examples follow.
		 *
		 *  To determine if the 'AW_CAMERA_FLAG_TRACK_USER' flag is set:
		 *
		 *    bool bTrackUser = (cam.Flags() & AW_CAMERA_FLAG_TRACK_USER)? true : false;
		 *
		 *  To remove a flag:
		 *
		 *    unsigned short& opts = cam.Flags();
		 *    opts &= ~AW_CAMERA_FLAG_TRACK_USER;
		 *
		 *  To add a flag:
		 *
		 *    unsigned short& opts = cam.Flags();
		 *    opts |= AW_CAMERA_FLAG_TRACK_USER;
		 */
		unsigned short& Flags() { return m_data.flags; }

		/** @brief Determine the camera's zoom factor.
		 *
		 *  Enter 1 for a normal ninety-degree view. Lower values will
		 *  produce a wide-angle lens, while higher values will produce
		 *  a zoom effect.
		 */
		float& Zoom() { return m_data.zoom; }

		/** @brief Determine the camera's tag name
		 *
		 *  Use this to assign a name to the camera.
		 *  A camera must have a name to be used.
		 */
		SimpleString& TagName() { return m_tag; }

		/** @brief Create an aw_object_data_camera structure
		 *
		 *  Creates and returns an aw_object_data_camera structure
		 *  using the data from this wrapper class.
		 *
		 *  Note that it is up to the caller to appropriately use
		 *  free() to delete the memory associated with this
		 *  structure.
		 */
		aw_object_data_camera* GetObjectData();

		/** @brief Default constructor.
		 *
		 *  Constructor which initializes the camera wrapper class.
		 */
		Camera();

		/** @brief Assign camera data from an SDK structure.
		 *
		 *  Constructor which initializes the camera wrapper class from
		 *  an existing aw_object_data_camera structure.
		 */
		void Assign(const aw_object_data_camera* data);

	protected:
		aw_object_data_camera m_data;
		SimpleString m_tag;

		Camera(const Camera&);
	};

	/** @brief A wrapper class defining a 'mover' object.
	 *
	 * This class encapsulates a mover object. The methods return
	 * references to various fields, which may be used to update them.
	 *
	 * For example:
	 *
	 *     char& glide = mov.GlideFactor();
	 *
	 * Will allow you to update the mover's glide factor by assigning a
	 * value to the 'glide' reference.
	 */
	class Mover
	{
	public:
		/** @brief Determine this mover's type.
		 *
		 *  AW_MOVER_TYPE_PASSIVE: "Server Controlled"
		 *  AW_MOVER_TYPE_ACTIVE:  "User Controlled" (default)
		 *  AW_MOVER_TYPE_ITEM:    "Pick-Up Item"
		 */
		unsigned char& Type() { return m_data.type; }

		/** @brief Return the bitfield containing mover options.
		 *
		 *  This value is a bitfield which contains all of the active
		 *  mover options.
		 *
		 *  Mover flags are defined in AW_MOVER_FLAGS. Examples follow.
		 *
		 *  To determine if the 'AW_MOVER_FLAG_SINGLE_RIDER' flag is set:
		 *
		 *    bool bSingleRider = (mov.Flags() & AW_MOVER_FLAG_SINGLE_RIDER)? true : false;
		 *
		 *  To remove a flag:
		 *
		 *    unsigned int& opts = mov.Flags();
		 *    opts &= ~AW_MOVER_FLAG_SINGLE_RIDER;
		 *
		 *  To add a flag:
		 *
		 *    unsigned int& opts = mov.Flags();
		 *    opts |= AW_MOVER_FLAG_SINGLE_RIDER;
		 */
		unsigned int& Flags() { return m_data.flags; }

		/** @brief Determine X coordinate of this mover's locked position.
		 *
		 *  If 'Lock Av Pos' is enabled, this defines the X (West/East)
		 *  coordinate where the user will be locked.
		 */
		short& LockedPosX() { return m_data.locked_pos_x; }

		/** @brief Determine X coordinate of this mover's locked position.
		 *
		 *  If 'Lock Av Pos' is enabled, this defines the Y (height)
		 *  coordinate where the user will be locked.
		 */
		short& LockedPosY() { return m_data.locked_pos_y; }

		/** @brief Determine X coordinate of this mover's locked position.
		 *
		 *  If 'Lock Av Pos' is enabled, this defines the Z (North/South)
		 *  coordinate where the user will be locked.
		 */
		short& LockedPosZ() { return m_data.locked_pos_z; }

		/** @brief Determine yaw of this mover's locked position
		 *
		 *  If 'Lock Av Pos' is enabled, this defines the yaw (direction)
		 *  where the user will be locked.
		 */
		short& LockedYaw() { return m_data.locked_yaw; }

		/** @brief Determine pitch of this mover's locked position.
		 *
		 *  If 'Lock Av Pos' is enabled, this defines the avatar pitch
		 *  where the user will be locked.
		 */
		short& LockedPitch() { return m_data.locked_pitch; }

		/** @brief Determine this mover's glide factor.
		 *
		 *  This determines how much speed the mover will lose when not
		 *  colliding with the ground or other objects
		 */
		char& GlideFactor() { return m_data.glide_factor; }

		/** @brief Determine this mover's speed factor.
		 *
		 *  This controls how fast the mover will go when under
		 *  user control.
		 */
		unsigned char& SpeedFactor() { return m_data.speed_factor; }

		/** @brief Determine this mover's friction factor.
		 *
		 *  This determines how much speed the mover will lose when
		 *  colliding with the ground or other objects.
		 */
		unsigned char& FrictionFactor() { return m_data.friction_factor; }

		/** @brief Determine the X coordinate for 'Roll on Move'.
		 *
		 *  This is how much the vehicle will lean into (or out of) turns.
		 */
		char& AccelTiltX() { return m_data.accel_tilt_x; }

		/** @brief Determine the Z coordinate for 'Roll on Move'.
		 *
		 *  This is how much the vehicle will lean into (or out of) turns.
		 */
		char& AccelTiltZ() { return m_data.accel_tilt_z; }

		/** @brief Determine this mover's turn faxtor.
		 *
		 *  This controls how tightly the mover may turn when under
		 *  user control.
		 */
		unsigned char& TurnFactor() { return m_data.turn_factor; }

		/** @brief Determine this pick-up item's avatar tag.
		 *
		 *  If this mover is a pick-up item, this defines the tag where
		 *  the item will appear on the user's avatar.
		 */
		unsigned char& AvatarTag() { return m_data.avatar_tag; }

		/** @brief Determine this mover's name */
		SimpleString& Name() { return m_name; }

		/** @brief Determine this mover's SEQ file */
		SimpleString& SEQ() { return m_seq; }

		/** @brief Determine this mover's script (unused???) */
		SimpleString& Script() { return m_script; }

		/** @brief Determine this mover's sound effect */
		SimpleString& Sound() { return m_sound; }

		/** @brief Determine this mover's bump object name */
		SimpleString& BumpName() { return m_bumpname; }

		/** @brief Class encapsulating a mover waypoint
		 *
		 *  This waypoint encapsulation is a simple linked list of
		 *  waypoints.
		 */
		class Waypoint
		{
		friend class Mover;
		public:
			/** Retrieve the waypoint's X (West/East) coordinate */
			short& X() { return m_data.x; }

			/** Retrieve the waypoint's Y (height) coordinate */
			short& Y() { return m_data.z; }

			/** Retrieve the waypoint's Z (North/South) coordinate */
			short& Z() { return m_data.z; }

			/** Retrieve the waypoint's direction (yaw) */
			short& Yaw() { return m_data.yaw; }

			/** Retrieve the waypoint's pitch */
			short& Pitch() { return m_data.pitch; }

			/** Retrieve the waypoint's roll */
			short& Roll() { return m_data.roll; }

			/** @brief Retrieve the waypoint's speed
			 *
			 *  This is how fast the mover will travel TO this waypoint.
			 */
			unsigned short& Speed() { return m_data.speed; }

			/** @brief Retrieve the waypoint's pause time
			 *
			 *  This is how long the mover will stay at this waypoint
			 *  before going to the next one.
			 */
			unsigned char& Pause() { return m_data.pause; }

			/** @brief Retrieve the waypoint's 'flag' field
			 */
			unsigned short& Flag() { return m_data.flag; }

			/** Default constructor */
			Waypoint(short wpX, short wpY, short wpZ, short wpYaw, short wpPitch, short wpRoll, unsigned short wpSpeed, unsigned char wpPause)
				: m_next(NULL)
			{
				m_data.x = wpX;
				m_data.y = wpY;
				m_data.z = wpZ;
				m_data.yaw = wpYaw;
				m_data.pitch = wpPitch;
				m_data.roll = wpRoll;
				m_data.speed = wpSpeed;
				m_data.pause = wpPause;
				m_data.flag = 0; // what is this, anyway ???
			}

			/** Constructor which accepts an SDK structure */
			Waypoint(const aw_waypoint& data)
				: m_next(NULL)
			{
				memcpy(&m_data, &data, sizeof(aw_waypoint));
			}

			/** Retrieve the next waypoint in the list */
			Waypoint* GetNext() const { return m_next; }

			/** Assign this waypoint's data to an SDK structure */
			void Assign(aw_waypoint* recv) { memmove(recv, &m_data, sizeof(aw_waypoint)); }

		protected:
			aw_waypoint m_data;
			Waypoint* m_next;

			/** Prevent copy-construction */
			Waypoint(const Waypoint&);
		};

		/** Add waypoint to this mover. */
		void AddWaypoint(Waypoint* wp);

		/** @brief Remove waypoint from this mover.
		 *
		 *  For safety, this requires a reference to the waypoint
		 *  which will be assigned NULL upon removal.
		 */
		void RemoveWaypoint(Waypoint*& wp);

		/** Remove all waypoints from this mover. */
		void ClearWaypoints();

		/** Retrieve the first waypoint */
		Waypoint* FirstWaypoint() { return m_waypoint; }

		/** @brief Create an aw_object_data_camera structure
		 *
		 *  Creates and returns an aw_object_data_camera structure
		 *  using the data from this wrapper class.
		 *
		 *  Note that it is up to the caller to appropriately use
		 *  free() to delete the memory associated with this
		 *  structure.
		 */
		aw_object_data_mover* GetObjectData();

		/** @brief Default constructor.
		 *
		 *  Constructor which initializes the camera wrapper class.
		 */
		Mover();

		/** @brief Assign mover data from an SDK structure.
		 *
		 *  Method which initializes the mover wrapper class from
		 *  an existing aw_object_data_mover structure.
		 */
		void Assign(const aw_object_data_mover* recv);

	protected:
		aw_object_data_mover m_data;
		SimpleString m_name;
		SimpleString m_seq;
		SimpleString m_script;
		SimpleString m_sound;
		Waypoint* m_waypoint;
		SimpleString m_bumpname;
		size_t m_wpcount;

		Mover(const Mover&);
	};

	/** @brief Prepares a zone object to be built in-world
	 *
	 * After calling this method use either the aw_object_add or
	 * aw_object_load SDK methods, after setting additional attributes
	 * as necessary.
	 *
	 * @param zone_p A pointer to the zone to be built.
	 */
	 void PrepareToBuild(aw_object_data_zone* zone_p);

	 /** @brief Prepares an emitter object to be built in-world
	  *
	  * After calling this method use either the aw_object_add or
	  * aw_object_load SDK methods, after setting additional attributes
	  * as necessary.
	  *
	  * @param particle_p A pointer to the emitter to be built.
	  */
	 void PrepareToBuild(aw_object_data_particles* particle_p);

	 /** @brief Prepares a camera object to be built in-world
	  *
	  * After calling this method use either the aw_object_add or
	  * aw_object_load SDK methods, after setting additional attributes
	  * as necessary.
	  *
	  * @param camera_p A pointer to the camera to be built.
	  */
	 void PrepareToBuild(aw_object_data_camera* camera_p);

	 /** @brief Prepares a mover object to be built in-world
	  *
	  * After calling this method use either the aw_object_add or
	  * aw_object_load SDK methods, after setting additional attributes
	  * as necessary.
	  *
	  * @param mover_p A pointer to the mover to be built.
	  */
	 void PrepareToBuild(aw_object_data_mover* mover_p);
};

/*---------------------------------------------------------------------------*/
#endif /*_AWV4_H__*/
